# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0016_auto_20150819_1900'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='start_datetime',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



import json

from base64 import b64decode
from beanstalk_dispatch.execution import execute_function
from jsonview.decorators import json_view
from jsonview.exceptions import BadRequest
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

import logging
logger = logging.getLogger(__name__)


@require_http_methods(['POST'])
@json_view
@csrf_exempt
def dispatcher(request):
    function_request = json.loads(b64decode(request.body.decode()).decode())
    try:
        execute_function(function_request)
        return {}
    except Exception as e:
        logger.error('Failure running function', exc_info=True)
        raise BadRequest(e)



from enum import Enum

from django.db import models
from django.db.models import CASCADE
from django.utils import timezone


class DeleteMixin(object):
    """
    Overrides delete and sets `is_deleted=True` instead of deleting object.
    Intended to be used with models that have an `is_deleted` field.
    """
    def delete(self, actually_delete=False, *args, **kwargs):
        # Implement cascading deletes.
        # NOTE(lydia): This ignores any other constraints passed to on_delete.
        if actually_delete:
            return super().delete(*args, **kwargs)
        fields = self._meta.get_fields(include_hidden=True)
        for field in fields:
            # Pass on fields that are not many-to-one relationships.
            if not ((field.one_to_many or field.one_to_one) and
                    field.auto_created):
                continue

            # Check if foreign key is set up with on_delete=CASCADE.
            if field.on_delete != CASCADE:
                continue

            # Query for all related objects and delete them.
            query_args = {field.field.name: self.id}
            related_objs = field.related_model.objects.filter(
                **query_args)
            for obj in related_objs:
                obj.delete()

        # Mark current object as deleted.
        self.is_deleted = True
        self.save()


class BaseModelManager(models.Manager):
    """
    Model manager intended to be used with models with an `is_deleted` field.
    Overrides the initial QuerySet to filter for objects where `is_deleted`
    is false.
    """

    def get_queryset(self):
        return (super(BaseModelManager, self).get_queryset()
                .filter(is_deleted=False))


class BaseModel(DeleteMixin, models.Model):
    """
    Abstract base class models which defines created_at and is_deleted fields.

    Attributes:
        created_at (datetime.datetime):
            Datetime at which the model is created.
        is_deleted (boolean):
            If value is True, mdoel is deleted. Default is False.
    """
    objects = BaseModelManager()
    unsafe_objects = models.Manager()

    created_at = models.DateTimeField(default=timezone.now)
    is_deleted = models.BooleanField(default=False)

    class Meta:
        abstract = True


class ChoicesEnum(Enum):
    """An enum suitable for use in Django model field choices.

    Allows defining enums using descriptive strings, then automatically assigns
    integer values to them for storage in an `IntegerField.` Example usage:

    >>> class MyEnum(ChoicesEnum):
    ...     item0 = 'My first item'
    ...     item1 = 'My other item'
    ...     blue = 'My third item'
    ...
    >>> class MyModel(models.Model):
    ...     my_field = models.IntegerField(choices=MyEnum.choices())
    ...
    >>> m = MyModel(my_field=MyEnum.blue.value)
    >>> m.my_field
    2
    >>> m.get_my_field_display()
    'My third item'
    """
    def __new__(cls, description):
        """ Assigns 0-indexed values to each enum member on class creation. """
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj.description = description
        obj._value_ = value
        return obj

    def as_choice(self):
        """ Formats an enum member for use in a Django model field choice. """
        return (self.value, self.description)

    def as_dict(self):
        return {
            'value': self.value,
            'description': self.description
        }

    def __repr__(self):
        return "<{}.{}: {} ({})>".format(
            self.__class__.__name__,
            self.name,
            self.value,
            self.description)

    @classmethod
    def choices(cls):
        """ Formats all enum members for use in Django model field choices.

        Output looks like:
        >>> MyEnum.choices()
        [
            (0, 'My first item'),
            (1, 'My other item'),
            (2, 'My third item'),
        ]

        And it can be used like:
        >>> some_field = IntegerField(choices=MyEnum.choices())

        """
        # Ideally this would be a generator, but Django seems to break when
        # passing choices=<generator> into a model field.
        return [item.as_choice() for item in cls]

    @classmethod
    def serialize(cls):
        """ Formats all enum members for external use (e.g., angular code).

        Output looks like:
        >>> MyEnum.serialize()
        {
            'MY_FIRST_ITEM': {
                'value': 0,
                'description': 'My first name'
            },
            ...
        }
        """
        return {item.name: item.as_dict() for item in cls}



import json

from base64 import b64encode
from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS
from beanstalk_dispatch.common import create_request_body
from django.core.urlresolvers import reverse
from django.test import Client
from django.test import TestCase
from django.test import override_settings

CALL_COUNTER = 0


def counter_incrementer(first_arg, second_arg=None):
    global CALL_COUNTER
    CALL_COUNTER += first_arg
    if second_arg:
        CALL_COUNTER += second_arg


DISPATCH_SETTINGS = {
    'BEANSTALK_DISPATCH_TABLE': {
        'the_counter': ('beanstalk_dispatch.tests.'
                        'test_dispatcher.counter_incrementer')
    }
}


class DispatcherTestCase(TestCase):

    """ Test the server-side function dispatcher.

    In these tests, we base64-encode every message we send to the server
    because this is what boto does.
    """

    def setUp(self):
        global CALL_COUNTER
        CALL_COUNTER = 0
        self.client = Client()
        self.url = reverse('beanstalk_dispatcher')

    @override_settings(**DISPATCH_SETTINGS)
    def test_no_get(self):
        response = self.client.get(self.url)
        self.assertEquals(response.status_code, 405)

    @override_settings(BEANSTALK_DISPATCH_TABLE=None)
    def test_no_dispatch(self):
        response = self.client.post(
            self.url, b64encode(
                create_request_body('some_func').encode('ascii')),
            content_type='application/json')
        self.assertEquals(response.status_code, 400)
        self.assertEquals(json.loads(response.content.decode()),
                          {'message': 'No beanstalk dispatch table configured',
                           'error': 400})

    @override_settings(**DISPATCH_SETTINGS)
    def test_missing_function(self):
        response = self.client.post(
            self.url,
            b64encode(create_request_body('nonexistent_func').encode('ascii')),
            content_type='application/json')
        self.assertEquals(response.status_code, 400)
        self.assertEquals(
            json.loads(response.content.decode()),
            {'message': 'Requested function not found: nonexistent_func',
                'error': 400})

    @override_settings(**DISPATCH_SETTINGS)
    def test_malformed_request(self):
        keys = {FUNCTION, ARGS, KWARGS}
        for missing_key in keys:
            request_body = {key: 'test' for key in
                            keys - {missing_key}}
            response = self.client.post(
                self.url,
                b64encode(json.dumps(request_body).encode('ascii')),
                content_type='application/json')
            self.assertEquals(response.status_code, 400)
            self.assertEquals(json.loads(response.content.decode()), {
                'message': 'Please provide a {} argument'.format(missing_key),
                'error': 400})

    @override_settings(**DISPATCH_SETTINGS)
    def test_both_args_kwargs(self):
        body = b64encode(
            create_request_body('the_counter', 1, second_arg=5)
            .encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(json.loads(response.content.decode()),
                          {})
        self.assertEquals(CALL_COUNTER, 6)

    @override_settings(**DISPATCH_SETTINGS)
    def test_just_args(self):
        body = b64encode(create_request_body('the_counter', 2).encode('ascii'))
        response = self.client.post(self.url,
                                    body,
                                    content_type='application/json')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(json.loads(response.content.decode()),
                          {})
        self.assertEquals(CALL_COUNTER, 2)



from django.conf.urls import include
from django.conf.urls import url

from orchestra.views import dashboard_tasks
from orchestra.views import get_timer
from orchestra.views import new_task_assignment
from orchestra.views import save_task_assignment
from orchestra.views import start_timer
from orchestra.views import status
from orchestra.views import stop_timer
from orchestra.views import submit_task_assignment
from orchestra.views import task_assignment_information
from orchestra.views import upload_image
from orchestra.views import update_timer
from orchestra.views import TimeEntryList
from orchestra.views import TimeEntryDetail
from orchestra.project_api.views import assign_worker_to_task
from orchestra.project_api.views import create_project
from orchestra.project_api.views import workflow_types
from orchestra.project_api.views import project_details_url
from orchestra.project_api.views import project_information


urlpatterns = [
    # Interface API
    url(r'^interface/dashboard_tasks/$',
        dashboard_tasks, name='dashboard_tasks'),

    url(r'^interface/task_assignment_information/$',
        task_assignment_information,
        name='task_assignment_information'),

    url(r'^interface/save_task_assignment/$',
        save_task_assignment,
        name='save_task_assignment'),

    url(r'^interface/submit_task_assignment/$',
        submit_task_assignment,
        name='submit_task_assignment'),

    url(r'^interface/new_task_assignment/(?P<task_type>\w+)/$',
        new_task_assignment,
        name='new_task_assignment'),

    url(r'^interface/upload_image/$',
        upload_image,
        name='upload_image'),

    url(r'^interface/time_entries/$',
        TimeEntryList.as_view(), name='time_entries'),

    url(r'^interface/time_entries/(?P<pk>[0-9]+)/$',
        TimeEntryDetail.as_view(), name='time_entry'),

    url(r'^interface/timer/start/$', start_timer, name='start_timer'),
    url(r'^interface/timer/stop/$', stop_timer, name='stop_timer'),
    url(r'^interface/timer/$', get_timer, name='get_timer'),
    url(r'^interface/timer/update/$', update_timer, name='update_timer'),

    url(r'^interface/project_management/',
        include('orchestra.interface_api.project_management.urls',
                namespace='project_management')),

    # Client API
    url(r'^project/project_information/$',
        project_information,
        name='project_information'),
    url(r'^project/create_project/$',
        create_project,
        name='create_project'),
    url(r'^project/workflow_types/$',
        workflow_types,
        name='workflow_types'),
    url(r'^project/project_details_url/$',
        project_details_url,
        name='project_details_url'),
    url(r'^project/assign_worker_to_task/$',
        assign_worker_to_task,
        name='assign_worker_to_task'),
    url(r'^status/$',
        status,
        name='status'),
]



from unittest.mock import patch

from orchestra.core.errors import MachineExecutionError
from orchestra.machine_tasks import execute
from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.utils.task_lifecycle import create_subsequent_tasks


class MachineTaskTestCase(OrchestraTestCase):
    def setUp(self):
        super().setUp()
        setup_models(self)

        machine_workflow_version = (
            self.workflow_versions['machine_workflow_version'])
        self.step = machine_workflow_version.steps.get(slug='machine_step')
        self.assertIsNotNone(self.step)
        self.step.execution_function = {
            'path': 'orchestra.tests.helpers.workflow.machine_task_function',
        }
        self.step.save()
        patcher = patch(
            'orchestra.tests.helpers.workflow.machine_task_function',
            return_value={'data': ''})
        self.machine_function_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.project = ProjectFactory(
            workflow_version=machine_workflow_version)
        create_subsequent_tasks(self.project)
        self.task = self.project.tasks.first()
        self.assertIsNotNone(self.task)

        # Since create_subsequent_tasks will automatically run the machine task
        # we need to reset the task each time it's called
        self._reset_task()

    def _reset_task(self):
        self.task.status = Task.Status.AWAITING_PROCESSING
        self.task.save()
        self.task.assignments.all().delete()
        self.machine_function_mock.call_count = 0

    def _assert_correct_machine_task_state(
            self, assignment_status, task_status):
        self.task.refresh_from_db()
        assignment = self.task.assignments.first()
        self.assertEquals(assignment.status, assignment_status)
        self.assertEquals(self.task.status, task_status)

        # Check correct iteration state
        self.assertEquals(assignment.iterations.count(), 1)
        iteration = assignment.iterations.first()
        if task_status == Task.Status.COMPLETE:
            expected_status = Iteration.Status.REQUESTED_REVIEW
            expected_data = assignment.in_progress_task_data
            self.assertIsNotNone(iteration.end_datetime)
        else:
            expected_status = Iteration.Status.PROCESSING
            expected_data = {}
            self.assertIsNone(iteration.end_datetime)

        self.assertEquals(iteration.status, expected_status)
        self.assertEquals(iteration.submitted_data, expected_data)

        # Assert that machine function is called once and there is only one
        # assignment regardless of state
        self.assertEquals(self.machine_function_mock.call_count, 1)
        self.assertEquals(self.task.assignments.count(), 1)

    def test_new_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    def test_already_processing_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)

        # Pretend that the task is still processing
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        assignment = self.task.assignments.first()
        assignment.status = TaskAssignment.Status.PROCESSING
        assignment.save()

        # Reset iteration for assignment
        assignment.iterations.all().delete()
        Iteration.objects.create(
            assignment=assignment,
            start_datetime=assignment.start_datetime)

        # Another machine attempts to perform task
        with self.assertRaises(MachineExecutionError):
            execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.PROCESSING, Task.Status.PROCESSING)

    def test_already_completed_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)
        with self.assertRaises(MachineExecutionError):
            execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    @patch('orchestra.machine_tasks.logger')
    def test_marking_failed_task_assignment(self, mock_logger):
        self.machine_function_mock.side_effect = Exception('Function failed.')
        self._reset_task()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.FAILED, Task.Status.PROCESSING)
        mock_logger.exception.assert_called_with('Machine task has failed')
        self.machine_function_mock.side_effect = None

    def test_reassigning_failed_task_assignment(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)

        # Pretend that the machine task failed
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        assignment = self.task.assignments.first()
        assignment.status = TaskAssignment.Status.FAILED
        assignment.save()
        self.assertEquals(self.machine_function_mock.call_count, 1)
        self.machine_function_mock.call_count = 0

        # New machine picks up the failed task
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    def test_aborted_project(self):
        self._reset_task()
        self.project.status = Project.Status.ABORTED
        self.project.save()
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.ABORTED)



from unittest.mock import patch

from django.conf import settings
from django.test import override_settings

from orchestra.models import Iteration
from orchestra.models import Task
from orchestra.communication.slack import _project_slack_group_name
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import end_project
from orchestra.utils.task_lifecycle import submit_task


class BasicNotificationsTestCase(OrchestraTestCase):
    """
    Test modular functions in the notifications module
    """

    def setUp(self):
        super().setUp()
        setup_models(self)

    @override_settings(ORCHESTRA_SLACK_INTERNAL_ENABLED=True)
    @override_settings(ORCHESTRA_SLACK_EXPERTS_ENABLED=True)
    def test_notify_status_change(self):
        project = self.projects['empty_project']
        internal_name = settings.SLACK_INTERNAL_NOTIFICATION_CHANNEL.strip('#')
        internal_groups = [
            group for group in self.slack.groups.list().body['groups']
            if group['name'] == internal_name]
        internal_group_id = internal_groups[0]['id']
        internal_slack_messages = self.slack.get_messages(internal_group_id)
        experts_slack_messages = self.slack.get_messages(
            project.slack_group_id)

        def _validate_slack_messages(message_stub):
            """
            Check that correct slack message was sent if API key present.
            """
            self.assertIn(message_stub, internal_slack_messages.pop())
            self.assertIn(message_stub, experts_slack_messages.pop())

        task = TaskFactory(project=project,
                           step=self.test_step,
                           status=Task.Status.AWAITING_PROCESSING)

        # Entry-level worker picks up task
        self.assertEquals(task.status, Task.Status.AWAITING_PROCESSING)
        task = assign_task(self.workers[0].id, task.id)
        self.assertTrue(task.is_worker_assigned(self.workers[0]))

        # Notification should be sent to entry-level worker
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[0].user.email)
        self.assertEquals(notification['subject'],
                          "You've been assigned to a new task!")

        _validate_slack_messages('Task has been picked up by a worker.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=True):
            # Entry-level worker submits task
            task = submit_task(task.id, {}, Iteration.Status.REQUESTED_REVIEW,
                               self.workers[0])

        self.assertEquals(task.status, Task.Status.PENDING_REVIEW)
        # Notification should be sent to entry-level worker
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[0].user.email)
        self.assertEquals(notification['subject'],
                          'Your task is under review!')

        _validate_slack_messages('Task is awaiting review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Reviewer picks up task
        task = assign_task(self.workers[1].id, task.id)
        self.assertEquals(task.status, Task.Status.REVIEWING)
        # No notification should be sent
        self.assertEquals(len(self.mail.inbox), 0)

        _validate_slack_messages('Task is under review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Reviewer rejects task
        task = submit_task(task.id, {}, Iteration.Status.PROVIDED_REVIEW,
                           self.workers[1])
        self.assertEquals(task.status, Task.Status.POST_REVIEW_PROCESSING)
        # Notification should be sent to original worker
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[0].user.email)
        self.assertEquals(notification['subject'],
                          'Your task has been returned')

        _validate_slack_messages('Task was returned by reviewer.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Entry-level worker resubmits task
        task = submit_task(task.id, {}, Iteration.Status.REQUESTED_REVIEW,
                           self.workers[0])
        self.assertEquals(task.status, Task.Status.REVIEWING)
        # Notification should be sent to reviewer
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[1].user.email)
        self.assertEquals(notification['subject'],
                          'A task is ready for re-review!')

        _validate_slack_messages('Task is under review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # First reviewer accepts task
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=True):
            task = submit_task(task.id, {}, Iteration.Status.REQUESTED_REVIEW,
                               self.workers[1])
        self.assertEquals(task.status, Task.Status.PENDING_REVIEW)
        # Notification should be sent to first reviewer
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[1].user.email)
        self.assertEquals(notification['subject'],
                          'Your task is under review!')

        _validate_slack_messages('Task is awaiting review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Second reviewer picks up task
        task = assign_task(self.workers[3].id, task.id)
        self.assertEquals(task.status, Task.Status.REVIEWING)
        # No notification should be sent
        self.assertEquals(len(self.mail.inbox), 0)

        _validate_slack_messages('Task is under review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Second reviewer rejects task
        task = submit_task(task.id, {}, Iteration.Status.PROVIDED_REVIEW,
                           self.workers[3])
        self.assertEquals(task.status, Task.Status.POST_REVIEW_PROCESSING)
        # Notification should be sent to first reviewer
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[1].user.email)
        self.assertEquals(notification['subject'],
                          'Your task has been returned')

        _validate_slack_messages('Task was returned by reviewer.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # First reviewer resubmits task
        task = submit_task(task.id, {}, Iteration.Status.REQUESTED_REVIEW,
                           self.workers[1])
        self.assertEquals(task.status, Task.Status.REVIEWING)
        # Notification should be sent to second reviewer
        self.assertEquals(len(self.mail.inbox), 1)
        notification = self.mail.inbox.pop()
        self.assertEquals(notification['recipient'],
                          self.workers[3].user.email)
        self.assertEquals(notification['subject'],
                          'A task is ready for re-review!')

        _validate_slack_messages('Task is under review.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # Second reviewer accepts task; task is complete
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            task = submit_task(task.id, {}, Iteration.Status.REQUESTED_REVIEW,
                               self.workers[3])
        self.assertEquals(task.status, Task.Status.COMPLETE)

        # Notification should be sent to all workers on task
        self.assertEquals(len(self.mail.inbox), 3)
        recipients = {mail['recipient'] for mail in self.mail.inbox}
        subjects = {mail['subject'] for mail in self.mail.inbox}
        self.assertEquals(recipients,
                          {self.workers[uid].user.email for uid in (0, 1, 3)})
        self.assertEquals(subjects, {'Task complete!'})
        self.mail.clear()

        _validate_slack_messages('Task has been completed.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

        # End project
        end_project(task.project.id)
        task = Task.objects.get(id=task.id)
        self.assertEquals(task.status, Task.Status.ABORTED)

        # Notification should be sent to all workers on task
        self.assertEquals(len(self.mail.inbox), 3)
        recipients = {mail['recipient'] for mail in self.mail.inbox}
        subjects = {mail['subject'] for mail in self.mail.inbox}
        self.assertEquals(recipients,
                          {self.workers[uid].user.email for uid in (0, 1, 3)})
        self.assertEquals(subjects,
                          {'A task you were working on has been ended'})
        self.mail.clear()

        for task in project.tasks.all():
            _validate_slack_messages('Task has been aborted.')
        self.assertEquals(len(internal_slack_messages), 0)
        self.assertEquals(len(experts_slack_messages), 0)

    def test_slack_group_id(self):
        def fake_random_string():
            fake_random_string.num_calls = (
                getattr(fake_random_string, 'num_calls', 0) + 1)
            return str(fake_random_string.num_calls)

        for group_name in ('ketchup-3-sales-1', 'bongo-bash723581-3',
                           'bongo-bash723581-4'):
            self.slack.groups.create(group_name)

        project = self.projects['reject_rev_proj']

        # Ensure slack group names are properly slugified and randomized.
        project.short_description = "Sally's Plumbing Emporium | Write Copy"
        group_name = _project_slack_group_name(project)
        self.assertTrue(group_name.startswith('sallys-plumbing'))
        self.assertEqual(len(group_name), 20)

        # Ensure slack group names are properly slugified and randomized.
        project.short_description = "Sally's Plumbingorama Emporium | Write"
        group_name = _project_slack_group_name(project)
        self.assertTrue(group_name.startswith('sallys-plumbingo'))
        self.assertEqual(len(group_name), 21)

        # Test that we create unique slack IDs if there are conflicts by
        # mocking the randomization logic to return deterministic results and
        # ensure that we don't repeat project IDs.
        patch_path = 'orchestra.communication.slack._random_string'
        with patch(patch_path, new=fake_random_string):
            for short_description, group_id in (
                    # Because the mock function's counter is at 1 and
                    # ketchup-3-sales-1 already exists, the channel gets set
                    # to ketchup-3-sales-2.
                    ('ketchup #3 Sales --:/ kitchen supplies',
                     'ketchup-3-sales-2'),
                    # Because the mock function's counter is at 3 and bongo-
                    # bash723581-3 / bongo-bash723581-4 already exist, the
                    # channel gets set to bongo-bash723581-5.
                    ('!!bongo bash*&#$(*$7235818923701842312',
                     'bongo-bash723581-5'),
                    ('shortie', 'shortie-6'),
                    ('what a treat!', 'what-a-treat-7'),
                    ('what a treat!', 'what-a-treat-8'),
                    ('what a treat!', 'what-a-treat-9'),
                    ("Sally's Plumbing Emporium", 'sallys-plumbing-10')):
                project.short_description = short_description
                self.assertEqual(_project_slack_group_name(project), group_id)



from unittest.mock import patch

from django.test import override_settings

from orchestra.google_apps.convenience import create_project_google_folder
from orchestra.google_apps.convenience import Service
from orchestra.models import Project
from orchestra.communication.slack import create_project_slack_group
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.google_apps import mock_create_drive_service
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import setup_models


class BasicTaskLifeCycleTestCase(OrchestraTestCase):
    """
    Test project module helper functions.
    """
    def setUp(self):  # noqa
        super().setUp()
        setup_models(self)

    @override_settings(GOOGLE_APPS=True)
    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_create_project_google_folder(self):
        project = self.projects['empty_project']
        # TODO(jrbotros): add additional functionality to google mock
        project_folder = create_project_google_folder(project)
        self.assertEquals(project_folder['id'], 1)
        self.assertEquals(
            project.team_messages_url, 'http://a.google.com/link')

    @override_settings(ORCHESTRA_SLACK_EXPERTS_ENABLED=True)
    def test_create_project_slack_group(self):
        groups = self.slack.data['groups']
        num_groups = len(groups)
        project = ProjectFactory(
            workflow_version=self.workflow_versions['test_workflow'])
        self.assertFalse(project.id in groups)
        group_id = create_project_slack_group(project)
        self.assertEquals(len(groups), num_groups + 1)
        self.assertTrue(group_id in groups)
        project = Project.objects.get(id=project.id)
        self.assertEquals(group_id, project.slack_group_id)



from django.db.models import Count
from orchestra.models import Project
from orchestra.models import Task


def completed_projects(projects):
    """
    Filters `projects` queryset to completed ones.

    TODO(marcua): if we use this for business logic, we should also ensure
    that the number of tasks equals the number of steps in the workflow, which
    will be easier after @thisisdhaas implements workflows-near-the-db.  It
    would be cleaner to just add a Complete state to Projects at that point.
    """
    projects = projects.filter(status=Project.Status.ACTIVE)
    in_progress_projects = (
        Task.objects.exclude(status=Task.Status.COMPLETE)
        .values_list('project', flat=True))
    return (
        projects.annotate(num_tasks=Count('tasks'))
        .filter(num_tasks__gt=0)
        .exclude(id__in=in_progress_projects))



from django.apps import apps as django_apps
from django.conf import settings
from django.db import transaction

from orchestra.models import Certification
from orchestra.models import Workflow
from orchestra.models import Step
from orchestra.models import WorkflowVersion
from orchestra.workflow.defaults import get_default_assignment_policy
from orchestra.workflow.defaults import get_default_review_policy
from orchestra.workflow.directory import parse_workflow_directory
from orchestra.core.errors import WorkflowError


def get_workflow_version_slugs():
    versions = {}
    for app_name in settings.ORCHESTRA_WORKFLOWS:
        # App label is the last part of the app name by default
        app_label = app_name.split('.')[-1]
        workflow_directory = django_apps.get_app_config(app_label).path
        data = parse_workflow_directory(workflow_directory)
        workflow_slug = data['workflow']['slug']
        if versions.get(workflow_slug) is not None:
            raise WorkflowError('Workflow {} present in multiple apps: {}, {}'
                                .format(workflow_slug,
                                        versions[workflow_slug]['app_label'],
                                        app_label))
        else:
            versions[workflow_slug] = {
                'app_label': app_label,
                'versions': (version['slug'] for version in data['versions'])
            }
    return versions


@transaction.atomic
def load_workflow(app_label, version_slug, force=False):
    workflow_directory = django_apps.get_app_config(app_label).path
    data = parse_workflow_directory(workflow_directory)

    # Create the workflow object if it doesn't exist
    workflow_data = data['workflow']
    workflow, workflow_created = Workflow.objects.update_or_create(
        slug=workflow_data['slug'],
        defaults={
            'name': workflow_data['name'],
            'description': workflow_data['description'],
            'code_directory': workflow_directory,
            'sample_data_load_function': workflow_data.get(
                'sample_data_load_function')
        }
    )

    # Create all certifications for the workflow
    for certification_data in workflow_data['certifications']:
        Certification.objects.update_or_create(
            slug=certification_data['slug'],
            workflow=workflow,
            defaults={
                'name': certification_data['name'],
                'description': certification_data['description'],
            }
        )

    # Create the certification dependencies once all certs are in the db
    # Allow updating these over time so that a workflow's certifications can
    # evolve. This means that the user is responsible for ensuring that all
    # workers have the proper certifications after updating these dependencies.
    for certification_data in workflow_data['certifications']:
        certification = Certification.objects.get(
            slug=certification_data['slug'],
            workflow=workflow
        )
        required_certification_slugs = certification_data.get(
            'required_certifications', [])
        required_certifications = Certification.objects.filter(
            workflow=workflow,
            slug__in=required_certification_slugs
        )
        if required_certifications.count() != len(
                required_certification_slugs):
            raise WorkflowError(
                'Certification {} requires non-existent certification.'
                .format(certification_data['slug']))
        certification.required_certifications = list(required_certifications)

    # Load the desired versions
    desired_versions = [version_data for version_data in data['versions']
                        if version_data['slug'] == version_slug]
    if len(desired_versions) != 1:
        raise WorkflowError('Invalid version requested: {}'
                            .format(version_slug))
    load_workflow_version(desired_versions[0], workflow, force=force)


def load_workflow_version(version_data, workflow, force=False):
    # Create the version object
    version, version_created = WorkflowVersion.objects.update_or_create(
        slug=version_data['slug'],
        workflow=workflow,
        defaults={
            'name': version_data['name'],
            'description': version_data['description']
        }
    )

    if not version_created:
        if not force:
            # It is safe to error out after modifying the DB because
            # all of this code is wrapped in a transaction by load_workflow.
            raise WorkflowError('Version {} already exists'
                                .format(version_data['slug']))

        # Check that the versions are safe to merge
        new_step_slugs = set(step['slug'] for step in version_data['steps'])
        old_step_slugs = set(
            Step.objects
            .filter(workflow_version=version)
            .values_list('slug', flat=True)
        )
        if new_step_slugs != old_step_slugs:
            raise WorkflowError('Even with --force, cannot change the steps '
                                'of a workflow. Drop and recreate the '
                                'database to reset, or create a new version '
                                'for your workflow.')

    # Create or update the version steps.
    old_creation_dependencies = {}
    old_submission_dependencies = {}
    for step_data in version_data['steps']:
        is_human = step_data.get('is_human', True)
        step, step_created = Step.objects.update_or_create(
            slug=step_data['slug'],
            workflow_version=version,
            defaults={
                'name': step_data['name'],
                'description': step_data['description'],
                'is_human': is_human,
                'detailed_description_function': step_data.get(
                    'detailed_description_function', {}),
                'execution_function': step_data.get('execution_function', {}),
                'assignment_policy': step_data.get(
                    'assignment_policy',
                    get_default_assignment_policy(is_human)),
                'review_policy': step_data.get(
                    'review_policy',
                    get_default_review_policy(is_human)),
                'user_interface': step_data.get('user_interface', {}),
            }
        )
        if not step_created:
            old_creation_dependencies[step_data['slug']] = set(
                step.creation_depends_on.values_list('slug', flat=True))
            old_submission_dependencies[step_data['slug']] = set(
                step.submission_depends_on.values_list('slug', flat=True))

        # Don't prevent updates to these, because we want to allow
        # certifications to evolve over the lifetime of a workflow.
        _set_step_dependencies(step, step_data, 'required_certifications',
                               Certification, workflow=workflow)

    # Set up step dependencies once the steps objects are in the DB.
    for step_data in version_data['steps']:
        step_slug = step_data['slug']
        step = Step.objects.get(
            slug=step_slug,
            workflow_version=version
        )

        # Set step creation dependencies.
        _verify_dependencies_not_updated(
            step_data,
            'creation_depends_on',
            old_creation_dependencies.get(step_slug)
        )
        _set_step_dependencies(step, step_data, 'creation_depends_on', Step,
                               workflow_version=version)

        # Set step submission dependencies.
        _verify_dependencies_not_updated(
            step_data,
            'submission_depends_on',
            old_submission_dependencies.get(step_slug)
        )
        _set_step_dependencies(step, step_data, 'submission_depends_on', Step,
                               workflow_version=version)


def _verify_dependencies_not_updated(step_data, dependency_attr,
                                     old_dependencies):
    new_dependencies = set(step_data.get(dependency_attr, []))
    if old_dependencies is not None and old_dependencies != new_dependencies:
        raise WorkflowError(
            'Even with --force, cannot change the topology of a workflow. '
            'Drop and recreate the database to reset, or create a new '
            'version for your workflow.')


def _set_step_dependencies(step, step_data, dependency_attr, dependency_model,
                           **model_filters):
    dependency_slugs = set(step_data.get(dependency_attr, []))
    dependencies = list(dependency_model.objects.filter(
        slug__in=dependency_slugs, **model_filters))
    if len(dependencies) != len(dependency_slugs):
        raise WorkflowError(
            '{}.{} contains a non-existent slug.'
            .format(step_data['slug'], dependency_attr))
    setattr(step, dependency_attr, dependencies)



from django.core.management.base import BaseCommand

from orchestra.core.errors import WorkflowError
from orchestra.workflow.load import get_workflow_version_slugs
from orchestra.workflow.load import load_workflow


class Command(BaseCommand):
    help = 'Loads a workflow version into the database'

    def add_arguments(self, parser):
        parser.add_argument(
            '--force', action='store_true', default=False,
            help=('If a workflow version already exists, update the existing '
                  'version instead of exiting.'))

    def handle(self, *args, **options):
        workflow_version_slugs = get_workflow_version_slugs()
        for workflow_slug, workflow_info in workflow_version_slugs.items():
            for version_slug in workflow_info['versions']:
                self.stdout.write('Loading {} {} from {}'.format(
                    workflow_slug, version_slug, workflow_info['app_label']))
                self._load_single_workflow(
                    workflow_info['app_label'], version_slug, options['force'])

    def _load_single_workflow(self, app_label, version_slug, force):
        try:
            load_workflow(app_label, version_slug, force=force)
        except WorkflowError as e:
            self.stderr.write(
                'An error occurred while loading the workflow: {}'
                .format(e))



from orchestra.models.communication.models import CommunicationPreference  # noqa
from orchestra.models.communication.models import StaffBotRequest  # noqa
from orchestra.models.communication.models import StaffingRequestInquiry  # noqa
from orchestra.models.communication.models import StaffingResponse  # noqa
from orchestra.models.core.models import Workflow  # noqa
from orchestra.models.core.models import WorkflowVersion  # noqa
from orchestra.models.core.models import Certification  # noqa
from orchestra.models.core.models import Step  # noqa
from orchestra.models.core.models import Worker  # noqa
from orchestra.models.core.models import WorkerCertification  # noqa
from orchestra.models.core.models import Project  # noqa
from orchestra.models.core.models import Task  # noqa
from orchestra.models.core.models import TaskAssignment  # noqa
from orchestra.models.core.models import Iteration  # noqa
from orchestra.models.core.models import TimeEntry  # noqa
from orchestra.models.core.models import TaskTimer  # noqa
from orchestra.models.core.models import PayRate  # noqa



import json
import logging

logger = logging.getLogger(__name__)


def load_encoded_json(encoded_data):
    """
    Args: encoded_data (str):
        Encoded JSON data
    Returns:
        data (dict):
            Dictionary of decoded JSON content
    """
    # TODO(joshblum): grep around the code base and use this function where we
    # can
    try:
        json_data = encoded_data.decode('utf-8')
        data = json.loads(json_data)
    except (AttributeError, ValueError, UnicodeError) as e:
        logger.warn('Json decode error {}'.format(str(e)))
        data = {}
    return data



from rest_framework import permissions

from orchestra.models import Worker


class IsAssociatedWorker(permissions.BasePermission):
    """
    Permission for objects with `worker` field. Checks if request.user matches
    worker on object.
    """

    def has_object_permission(self, request, view, obj):
        worker = Worker.objects.get(user=request.user)
        return obj.worker == worker



from setuptools import setup, find_packages

from release import get_version

version = get_version('orchestra')
setup(
    name='orchestra',
    version=version,
    description='A framework for building complex expert workflows.',
    author='Unlimited Labs, Inc.',
    author_email='hello@unlimitedlabs.com',
    license='Apache 2',
    url='https://github.com/unlimitedlabs/orchestra',
    download_url=(
        'https://github.com/unlimitedlabs/orchestra/tarball/v' + version),
    keywords=['crowdsourcing', 'workflows'],
    classifiers=[],
    packages=find_packages(),
    include_package_data=True,
)



from django.apps import AppConfig


class OrchestraAppConfig(AppConfig):
    name = 'orchestra'
    verbose_name = 'Orchestra'

    def ready(self):
        from orchestra.accounts import signals  # noqa



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0007_auto_20150507_0204'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certification',
            name='required_certifications',
            field=models.ManyToManyField(
                related_name='required_certifications_rel_+', to='orchestra.Certification'),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='taskassignment',
            unique_together=set([('task', 'assignment_counter')]),
        ),
    ]



from unittest.mock import patch

from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.project_properties import completed_projects
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import submit_task


class ProjectPropertiesTestCase(OrchestraTestCase):
    def setUp(self):  # noqa
        super().setUp()
        setup_models(self)

    def test_completed_projects(self):
        projects = Project.objects.all()
        initial_task = assign_task(self.workers[6].id,
                                   self.tasks['awaiting_processing'].id)
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(initial_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[6])
        self.assertEquals(completed_projects(projects).count(), 0)

        next_task = assign_task(
            self.workers[6].id,
            initial_task.project.tasks.order_by('-start_datetime')[0].id)
        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(next_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[6])
        self.assertEquals(completed_projects(projects).count(), 0)

        next_task = assign_task(
            self.workers[6].id,
            initial_task.project.tasks.order_by('-start_datetime')[0].id)

        with patch('orchestra.utils.task_lifecycle._is_review_needed',
                   return_value=False):
            initial_task = submit_task(next_task.id, {},
                                       Iteration.Status.REQUESTED_REVIEW,
                                       self.workers[6])
        self.assertEquals(completed_projects(projects).count(), 1)



from ajax_select import register
from ajax_select import LookupChannel
from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse
from django.utils.html import escape
from django.db.models import Q

from orchestra.models import CommunicationPreference
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import Worker

UserModel = get_user_model()


class FormatedItemMixin(object):
    url_reverse = ''

    def get_object_id(self, obj):
        return obj.id

    def format_item_display(self, obj):
        """ (HTML) formatted item for display in the dropdown """
        return '<div><a href="{}">{}</a></div>'.format(
            reverse(self.url_reverse, args=[self.get_object_id(obj)]),
            escape(obj)
        )


@register('users')
class UsersLookup(FormatedItemMixin, LookupChannel):

    model = UserModel
    url_reverse = 'admin:auth_user_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(username__icontains=q) |
            Q(first_name__icontains=q) |
            Q(last_name__icontains=q) |
            Q(email__icontains=q))


@register('workers')
class WorkerLookup(FormatedItemMixin, LookupChannel):

    model = Worker
    url_reverse = 'admin:orchestra_worker_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(user__first_name__icontains=q) |
            Q(user__last_name__icontains=q) |
            Q(user__email__icontains=q) |
            Q(user__username__icontains=q) |
            Q(slack_username__icontains=q))


@register('communication_preferences')
class CommunicationPreferenceLookup(FormatedItemMixin, LookupChannel):

    model = CommunicationPreference
    url_reverse = 'admin:orchestra_communicationpreference_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(worker__user__first_name__icontains=q) |
            Q(worker__user__last_name__icontains=q) |
            Q(worker__user__email__icontains=q) |
            Q(worker__user__username__icontains=q) |
            Q(worker__slack_username__icontains=q))


@register('task_assignments')
class TaskAssignmentLookup(FormatedItemMixin, LookupChannel):

    model = Task
    url_reverse = 'admin:orchestra_taskassignment_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(step__slug__icontains=q) |
            Q(project__short_description__icontains=q))


@register('tasks')
class TaskLookup(FormatedItemMixin, LookupChannel):

    model = Task
    url_reverse = 'admin:orchestra_task_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(step__slug__icontains=q) |
            Q(project__short_description__icontains=q))


@register('staffing_request_inquiries')
class StaffingRequestInquiryLookup(FormatedItemMixin, LookupChannel):

    model = StaffingRequestInquiry
    url_reverse = 'admin:orchestra_staffingrequestinquiry_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(
                communication_preference__worker__user__first_name__icontains=q
            ) |
            Q(communication_preference__worker__user__last_name__icontains=q) |
            Q(communication_preference__worker__user__email__icontains=q) |
            Q(communication_preference__worker__user__username__icontains=q) |
            Q(communication_preference__worker__slack_username__icontains=q) |
            Q(task__step__slug__icontains=q) |
            Q(task__project__short_description__icontains=q))


@register('projects')
class ProjectLookup(FormatedItemMixin, LookupChannel):

    model = Project
    url_reverse = 'admin:orchestra_project_change'

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(short_description__icontains=q) |
            Q(workflow_version__slug__icontains=q))
